@@ -27,6 +27,7 @@ from typing import Any, Dict, List, Optional, TYPE_CHECKING, Union, cast
 
 from .._dns import DNSAddress, DNSPointer, DNSQuestionType, DNSRecord, DNSService, DNSText
 from .._exceptions import BadTypeInNameException
+from .._logger import log
 from .._protocol.outgoing import DNSOutgoing
 from .._updates import RecordUpdate, RecordUpdateListener
 from .._utils.asyncio import get_running_loop, run_coro_with_timeout
@@ -124,19 +125,12 @@ class ServiceInfo(RecordUpdateListener):
         self.type = type_
         self._name = name
         self.key = name.lower()
+        self._ipv4_addresses: List[ipaddress.IPv4Address] = []
+        self._ipv6_addresses: List[ipaddress.IPv6Address] = []
         if addresses is not None:
-            self._addresses = addresses
+            self.addresses = addresses
         elif parsed_addresses is not None:
-            self._addresses = [_encode_address(a) for a in parsed_addresses]
-        else:
-            self._addresses = []
-        # This results in an ugly error when registering, better check now
-        invalid = [a for a in self._addresses if not isinstance(a, bytes) or len(a) not in (4, 16)]
-        if invalid:
-            raise TypeError(
-                'Addresses must be bytes, got %s. Hint: convert string addresses '
-                'with socket.inet_pton' % invalid
-            )
+            self.addresses = [_encode_address(a) for a in parsed_addresses]
         self.port = port
         self.weight = weight
         self.priority = priority
@@ -178,7 +172,21 @@ class ServiceInfo(RecordUpdateListener):
 
         This replaces all currently stored addresses, both IPv4 and IPv6.
         """
-        self._addresses = value
+        self._ipv4_addresses.clear()
+        self._ipv6_addresses.clear()
+
+        for address in value:
+            try:
+                addr = ipaddress.ip_address(address)
+            except ValueError:
+                raise TypeError(
+                    "Addresses must either be IPv4 or IPv6 strings, bytes, or integers;"
+                    f" got {address!r}. Hint: convert string addresses with socket.inet_pton"
+                )
+            if isinstance(addr, ipaddress.IPv4Address):
+                self._ipv4_addresses.append(addr)
+            else:
+                self._ipv6_addresses.append(addr)
 
     @property
     def properties(self) -> Dict:
@@ -194,10 +202,13 @@ class ServiceInfo(RecordUpdateListener):
     def addresses_by_version(self, version: IPVersion) -> List[bytes]:
         """List addresses matching IP version."""
         if version == IPVersion.V4Only:
-            return [addr for addr in self._addresses if not _is_v6_address(addr)]
+            return [addr.packed for addr in self._ipv4_addresses]
         if version == IPVersion.V6Only:
-            return list(filter(_is_v6_address, self._addresses))
-        return self._addresses
+            return [addr.packed for addr in self._ipv6_addresses]
+        return [
+            *(addr.packed for addr in self._ipv4_addresses),
+            *(addr.packed for addr in self._ipv6_addresses),
+        ]
 
     def parsed_addresses(self, version: IPVersion = IPVersion.All) -> List[str]:
         """List addresses in their parsed string form."""
@@ -220,7 +231,7 @@ class ServiceInfo(RecordUpdateListener):
 
         ll_addrs = list(filter(is_link_local, self.parsed_addresses(version)))
         other_addrs = list(filter(lambda addr: not is_link_local(addr), self.parsed_addresses(version)))
-        return ["{}%{}".format(addr, self.interface_index) for addr in ll_addrs] + other_addrs
+        return [f"{addr}%{self.interface_index}" for addr in ll_addrs] + other_addrs
 
     def _set_properties(self, properties: Dict) -> None:
         """Sets properties and text of this info from a dictionary"""
@@ -315,9 +326,20 @@ class ServiceInfo(RecordUpdateListener):
             return
 
         if isinstance(record, DNSAddress):
-            if record.key == self.server_key and record.address not in self._addresses:
-                self._addresses.append(record.address)
-                if record.type is _TYPE_AAAA and ipaddress.IPv6Address(record.address).is_link_local:
+            if record.key != self.server_key:
+                return
+            try:
+                ip_addr = ipaddress.ip_address(record.address)
+            except ValueError as ex:
+                log.warning("Encountered invalid address while processing %s: %s", record, ex)
+                return
+            if isinstance(ip_addr, ipaddress.IPv4Address):
+                if ip_addr not in self._ipv4_addresses:
+                    self._ipv4_addresses.insert(0, ip_addr)
+                return
+            if ip_addr not in self._ipv6_addresses:
+                self._ipv6_addresses.insert(0, ip_addr)
+                if ip_addr.is_link_local:
                     self.interface_index = record.scope_id
             return
 
@@ -422,13 +444,17 @@ class ServiceInfo(RecordUpdateListener):
     @property
     def _is_complete(self) -> bool:
         """The ServiceInfo has all expected properties."""
-        return not (self.text is None or not self._addresses)
+        return bool(self.text is not None and (self._ipv4_addresses or self._ipv6_addresses))
 
     def request(
         self, zc: 'Zeroconf', timeout: float, question_type: Optional[DNSQuestionType] = None
     ) -> bool:
         """Returns true if the service could be discovered on the
         network, and updates this object with details discovered.
+
+        While it is not expected during normal operation,
+        this function may raise EventLoopBlocked if the underlying
+        call to `async_request` cannot be completed.
         """
         assert zc.loop is not None and zc.loop.is_running()
         if zc.loop == get_running_loop():
@@ -494,10 +520,10 @@ class ServiceInfo(RecordUpdateListener):
 
     def __repr__(self) -> str:
         """String representation"""
-        return '%s(%s)' % (
+        return '{}({})'.format(
             type(self).__name__,
             ', '.join(
-                '%s=%r' % (name, getattr(self, name))
+                '{}={!r}'.format(name, getattr(self, name))
                 for name in (
                     'type',
                     'name',
